
#include <avr/io.h>
#include <avr/interrupt.h>

#include "keycode.h"
#include "queue16.h"
#include "ps2if.h"
#include "typedef.h"
#include "waitloop.h"

typedef struct Queue16 Queue;
#define qinit(Q) q16init(Q)
#define qpeek(Q) q16peek(Q)
#define qget(Q) q16get(Q)
#define qput(Q,C) q16put(Q,C)
#define qunget(Q,C) q16unget(Q,C)


#define NULL 0


Queue event_queue_l;
Queue event_queue_h;
Queue kb_input_queue;
Queue kb_output_queue;

unsigned short kb_input_bits;
unsigned short kb_output_bits;

unsigned short kb_timeout;



int countbits(unsigned short c)
{
	int i;
	for (i = 0; c; c >>= 1) {
		if (c & 1) i++;
	}
	return i;
}

// initializer

inline void initialize_kb_output()
{
	qinit(&kb_output_queue);
	kb_output_bits = 0;
}

inline void initialize_kb_input()
{
	qinit(&kb_input_queue);
	kb_input_bits = 0;
}




bool kb_next_output(unsigned char c)
{
	unsigned short d = c;
	if (!(countbits(d) & 1)) d |= 0x100;	// make odd parity
	d = (d | 0x600) << 1;
	//
	// d = 000011pdddddddd0
	//                    
	//                    ^ start bit
	//            ^^^^^^^^  data
	//           ^          odd parity
	//          ^           stop bit
	//         ^            reply from keyboard (ack bit)
	//
	cli();
	if (kb_input_bits) {
		sei();
		return false;
	}
	kb_output_bits = d;
	kb_set_clock_0();	// I/O inhibit, trigger interrupt
	kb_set_data_0();	// start bit
	sei();
	wait_40us();
	kb_set_clock_1();

	kb_timeout = 0;

	return true;
}

bool pc_send(unsigned char c)
{
	unsigned char i;
	unsigned short bits;

	bits = c & 0xff;
	if (!(countbits(bits) & 1)) bits |= 0x100;
	bits = (bits | 0x200) << 1;

	for (i = 0; i < 11; i++) {
		if (!pc_get_clock()) break;
		if (bits & 1)	pc_set_data_1();
		else			pc_set_data_0();
		bits >>= 1;
		wait_15us();
		if (!pc_get_clock()) break;
		pc_set_clock_0();
		wait_40us();
		pc_set_clock_1();
		wait_15us();
	}
	pc_set_data_1();

	return i == 11;
}

int pc_recv()
{
	int i;
	unsigned short bits;

	if (!pc_get_clock()) return -1;
	if (pc_get_data()) return -1;

	bits = 0;
	for (i = 0; i < 100; i++) {
		pc_set_clock_0();
		wait_40us();
		pc_set_clock_1();
		wait_15us();
		wait_15us();
		if (i < 9) {
			bits >>= 1;
			if (pc_get_data()) bits |= 0x100;
		} else {
			if (pc_get_data()) {	// stop bit ?
				break;
			}
		}
	}
	pc_set_clock_0();
	pc_set_data_0();
	wait_40us();
	pc_set_data_1();
	pc_set_clock_1();

	if (i != 9)
		return -1;							// framing error

	if (!(countbits(bits & 0x1ff) & 1))
		return -1;							// parity error

	return bits & 0xff;
}





inline void kb_put(unsigned char c)
{
	qput(&kb_output_queue, c & 0xff);
}

inline int kb_get()
{
	int c;
	cli();
	c = qget(&kb_input_queue);
	sei();
	return c;
}

inline unsigned short get_event()
{
	unsigned short l, h;
	cli();
	l = qget(&event_queue_l);
	h = qget(&event_queue_h);
	sei();
	if ((l | h) & 0xff00) return 0;
	return (h << 8) | l;
}

inline void put_event(unsigned short c)
{
	cli();
	qput(&event_queue_l, c);
	qput(&event_queue_h, c >> 8);
	sei();
}

inline void unget_event(unsigned short c)
{
	cli();
	qunget(&event_queue_l, c);
	qunget(&event_queue_h, c >> 8);
	sei();
}





SIGNAL(PCINT2_vect)
{
	if (!kb_get_clock()) {
		int c;

		kb_timeout = 10;	// 10ms

		if (!kb_input_bits) {
			if (kb_output_bits) {			// transmit mode
				if (kb_output_bits == 1) {
					kb_output_bits = 0;		// end transmit
					kb_timeout = 0;
				} else {
					if (kb_output_bits & 1) {
						kb_set_data_1();
					} else {
						kb_set_data_0();
					}
					kb_output_bits >>= 1;
				}
			} else {
				kb_input_bits = 0x800;		// start receive
			}
		}
		if (kb_input_bits) {
			kb_input_bits >>= 1;
			if (kb_get_data()) {
				kb_input_bits |= 0x800;
			}
			if (kb_input_bits & 1) {
				if (kb_input_bits & 0x800) {				// stop bit ?
					if (countbits(kb_input_bits & 0x7fc) & 1) {	// odd parity ?
						c = (kb_input_bits >> 2) & 0xff;
						qput(&kb_input_queue, c);
					}
				}
				kb_input_bits = 0;
				kb_timeout = 0;
			}
		}
	}
}


bool timerevent = false;

#define CLOCK 8000000UL
#define SCALE (CLOCK / 256)
#define PRESCALE 8

static unsigned short _scale = 0;

//SIGNAL(TIMER0_OVF_vect)
ISR(TIMER0_OVF_vect, ISR_NOBLOCK) // for 5576-003 problem
{
	_scale += 1000 * PRESCALE;
	if (_scale >= SCALE) {
		_scale -= SCALE;
		timerevent = true;
	}
}


void system_handler()
{
	int c;
	// transmit to keyboard
	c = qget(&kb_output_queue);
	if (c >= 0) {
		if (!kb_next_output(c)) {
			qunget(&kb_output_queue, c);
		}
	}
}





enum {
	PC_INIT			= 0x0000,
	PC_NORMAL		= 0x0100,
	PC_ED			= 0x0200,
	PC_F0			= 0x0300,
	PC_F3			= 0x0400,
	PC_FB			= 0x0500,
	PC_FC			= 0x0600,
	PC_FD			= 0x0700,
	PC_FF			= 0x0800,
};

enum {
	KB_INIT			= 0x0000,
	KB_NORMAL		= 0x0100,
	KB_WAIT			= 0x0200,
	KB_ED			= 0x0300,
	KB_F3			= 0x0400,
};

enum {
	EVENT_NULL					= 0,
	EVENT_INIT					= 0x0100,
	EVENT_KEYMAKE				= 0x0200,
	EVENT_KEYBREAK				= 0x0300,
	EVENT_SEND_KB_INDICATOR		= 0x0400,
	EVENT_SEND_KB_TYPEMATIC		= 0x0500,
};



unsigned char key_decode_state;
unsigned short kb_state;
unsigned char shiftflags;
unsigned char indicator;

unsigned short real_keyboard_id;
unsigned short fake_keyboard_id;
unsigned short _keyboard_id;
bool break_enable;
bool scan_enable;

unsigned char pspad_state;
unsigned char pspad_buttons0;
unsigned char pspad_buttons1;



void main_handler()
{
	int c;
	long event;

	if (timerevent) {
		timerevent = false;
		cli();
		if (kb_timeout > 0) {
			if (kb_timeout > 1) {
				kb_timeout--;
			} else {
				kb_output_bits = 0;
				kb_input_bits = 0;
				kb_set_data_1();
				kb_set_clock_1();
				kb_timeout = 0;
			}
		}
		sei();
	}

	event = get_event();
	if (event > 0) {
		switch (event & 0xff00) {
		case EVENT_INIT:
			if ((kb_state & 0xff00) != KB_INIT) {
				kb_put(0xff);
				kb_state = KB_INIT;
			}
			break;
		case EVENT_KEYMAKE:
			c = event & 0xff;
			switch (c) {
			case 44:	shiftflags |= KEYFLAG_SHIFT_L;		break;
			case 57:	shiftflags |= KEYFLAG_SHIFT_R;		break;
			case 58:	shiftflags |= KEYFLAG_CTRL_L;		break;
			case 64:	shiftflags |= KEYFLAG_CTRL_R;		break;
			case 60:	shiftflags |= KEYFLAG_ALT_L;		break;
			case 62:	shiftflags |= KEYFLAG_ALT_R;		break;
			case 112: // F1
			case 120: // F9
				pspad_buttons0 &= ~0x01; // select
				break;
			case 115: // F4
			case 123: // F12
				pspad_buttons0 &= ~0x08; // start
				break;
			case 31: // A
			case 79: // left
			case 92: // numpad 4
				pspad_buttons0 &= ~0x80; // left
				break;
			case 32: // S
			case 84: // down
			case 98: // numpad 2
				pspad_buttons0 &= ~0x40; // down
				break;
			case 33: // D
			case 89: // right
			case 102: // numpad 6
				pspad_buttons0 &= ~0x20; // right
				break;
			case 18: // W
			case 83: // up
			case 96: // numpad 8
				pspad_buttons0 &= ~0x10; // up
				break;
			case 36: // H
				pspad_buttons1 &= ~0x80; // square
				break;
			case 37: // J
				pspad_buttons1 &= ~0x40; // cross
				break;
			case 38: // K
				pspad_buttons1 &= ~0x20; // circle
				break;
			case 23: // U
				pspad_buttons1 &= ~0x10; // triangle
				break;
			case 48: // C
			case 53: // ,
				pspad_buttons1 &= ~0x04; // L1
				break;
			case 47: // X
			case 52: // M
				pspad_buttons1 &= ~0x01; // L2
				break;
			case 50: // B
			case 76: // Delete
			case 99: // numpad 0
				pspad_buttons1 &= ~0x08; // R1
				break;
			case 51: // N
			case 75: // Insert
			case 93: // numpad 1
				pspad_buttons1 &= ~0x02; // R2
				break;
			}
			break;
		case EVENT_KEYBREAK:
			c = event & 0xff;
			switch (c) {
			case 44:	shiftflags &= ~KEYFLAG_SHIFT_L;		break;
			case 57:	shiftflags &= ~KEYFLAG_SHIFT_R;		break;
			case 58:	shiftflags &= ~KEYFLAG_CTRL_L;		break;
			case 64:	shiftflags &= ~KEYFLAG_CTRL_R;		break;
			case 60:	shiftflags &= ~KEYFLAG_ALT_L;		break;
			case 62:	shiftflags &= ~KEYFLAG_ALT_R;		break;
			case 112: // F1
			case 120: // F9
				pspad_buttons0 |= 0x01; // select
				break;
			case 115: // F4
			case 123: // F12
				pspad_buttons0 |= 0x08; // start
				break;
			case 31: // A
			case 79: // left
			case 92: // numpad 4
				pspad_buttons0 |= 0x80; // left
				break;
			case 32: // S
			case 84: // down
			case 98: // numpad 2
				pspad_buttons0 |= 0x40; // down
				break;
			case 33: // D
			case 89: // right
			case 102: // numpad 6
				pspad_buttons0 |= 0x20; // right
				break;
			case 18: // W
			case 83: // up
			case 96: // numpad 8
				pspad_buttons0 |= 0x10; // up
				break;
			case 36: // H
				pspad_buttons1 |= 0x80; // square
				break;
			case 37: // J
				pspad_buttons1 |= 0x40; // cross
				break;
			case 38: // K
				pspad_buttons1 |= 0x20; // circle
				break;
			case 23: // U
				pspad_buttons1 |= 0x10; // triangle
				break;
			case 48: // C
			case 53: // ,
				pspad_buttons1 |= 0x04; // L1
				break;
			case 47: // X
			case 52: // M
				pspad_buttons1 |= 0x01; // L2
				break;
			case 50: // B
			case 76: // Delete
			case 99: // numpad 0
				pspad_buttons1 |= 0x08; // R1
				break;
			case 51: // N
			case 75: // Insert
			case 93: // numpad 1
				pspad_buttons1 |= 0x02; // R2
				break;
			}
			break;
		case EVENT_SEND_KB_INDICATOR:
			if (kb_state != KB_NORMAL) {
				unget_event(event);
			} else {
				kb_put(0xed);
				kb_state = KB_ED;
			}
			break;
		case EVENT_SEND_KB_TYPEMATIC:
			if (kb_state != KB_NORMAL) {
				unget_event(event);
			} else {
				kb_put(0xf3);
				kb_state = KB_F3;
			}
			break;
		}
	}

	c = kb_get();
	if (c >= 0) {
		unsigned short t;

		switch (kb_state) {
		case KB_INIT + 0:
			if (c == 0xfa) {
				c = -1;
			} if (c == 0xaa) {
				kb_put(0xf2);
				kb_state++;
				c = -1;
			}
			break;
		case KB_INIT + 1:
			if (c == 0xfa) {
				kb_state++;
				c = -1;
			}
			break;
		case KB_INIT + 2:	// receive keyboard ID 1st byte
			_keyboard_id = c;
			kb_state++;
			c = -1;
			break;
		case KB_INIT + 3:	// receive keyboard ID 2nd byte
			_keyboard_id |= c << 8;
			real_keyboard_id = _keyboard_id;
			kb_put(0xf3);
			kb_state++;
			c = -1;
			break;
		case KB_INIT + 4:
			if (c == 0xfa) {
				kb_put(0x00);
				kb_state++;
				c = -1;
			}
			break;
		case KB_INIT + 5:
			if (c == 0xfa) {
				kb_put(0xf4);
				kb_state++;
				c = -1;
			}
			break;
		case KB_INIT + 6:
			if (c == 0xfa) {
				kb_put(0xf0);
				kb_state++;
				c = -1;
			}
			break;
		case KB_INIT + 7:
			if (c == 0xfa) {
				switch (real_keyboard_id) {
				case 0x92ab:	// is IBM5576-001 ?
				case 0x90ab:	// is IBM5576-002 ?
				case 0x91ab:	// is IBM5576-003 ?
					kb_put(0x82);
					break;
				default:
					kb_put(0x02);
					break;
				}
				c = -1;
			}
			kb_state = KB_WAIT;
			break;
		case KB_ED:
			if (c == 0xfa) {
				kb_put(indicator);
				kb_state = KB_WAIT;
				c = -1;
			} else {
				kb_state = KB_NORMAL;
			}
			break;
		case KB_F3:
			if (c == 0xfa) {
				kb_put(0x23);
				kb_state = KB_WAIT;
				c = -1;
			} else {
				kb_state = KB_NORMAL;
			}
			break;
		case KB_WAIT:
			kb_state = KB_NORMAL;
			c = -1;
			break;
		}

		switch (c) {
		case -1:
			break;
		case 0xaa:
			kb_put(0xf2);
			kb_state = KB_INIT + 1;
			break;
		case 0xfa:
		case 0xee:
		case 0xfc:
		case 0xfe:
		case 0xff:
			key_decode_state = 0;
			kb_state = KB_NORMAL;
			break;
		default:
			switch (real_keyboard_id) {
			case 0x92ab:	// is IBM5576-001 ?
				t = ps2decode001(&key_decode_state, c);
				break;
			case 0x90ab:	// is IBM5576-002 ?
			case 0x91ab:	// is IBM5576-003 ?
				t = ps2decode002(&key_decode_state, c);
				break;
			default:
				t = ps2decode(&key_decode_state, c);
				break;
			}
			if (t && scan_enable) {
				unsigned short event;
				if (t & 0x8000) {
					event = (t & 0xff) | EVENT_KEYBREAK;
					put_event(event);
				} else {
					event = (t & 0xff) | EVENT_KEYMAKE;
					put_event(event);
				}
			}
			break;
		}
	}
}


void pspad_send(unsigned char c)
{
	SPDR = c;
	PORTB = 0xfd;	// ack = L
	waitloop(2);
	PORTB = 0xff;	// ack = H
}

void pspad_reset_state()
{
	DDRB = 0x00;
	PORTB = 0x00;
	SPCR = 0xac;	// disable SPI
	pspad_state = 0;
}

SIGNAL(PCINT0_vect)
{
	if (PINB & 0x04) {
		pspad_reset_state();
	} else {
		SPCR = 0xec;	// enable SPI
		PORTB = 0x13;
		DDRB = 0x13;
		pspad_state = 1;
	}
}

SIGNAL(SPI_STC_vect)
{
	int c = SPDR;
	sei();
	switch (pspad_state) {
	case 1:
		if (c == 0x01) {
			pspad_send(0x41);
			pspad_state++;
		} else {
			pspad_reset_state();
		}
		break;
	case 2:
		if (c == 0x42) {
			pspad_send(0x5a);
			pspad_state++;
		} else {
			pspad_reset_state();
		}
		break;
	case 3:
		pspad_send(pspad_buttons0);
		pspad_state++;
		break;
	case 4:
		pspad_send(pspad_buttons1);
		pspad_state = 0;
		break;
	default:
		pspad_reset_state();
		break;
	}
}

int main()
{
	int c;
	unsigned short t;

	DDRB = 0;
	PORTB = 0;

	PORTD = 0;
	DDRD = 0xaa;

	// enable pin change interrupt
	PCICR = 0x05;
	PCMSK0 = 0x04;
	PCMSK2 = 0x10;

	// enable interval timer
	TCCR0B = 0x02; // 1/8 prescaling
	TIMSK0 |= 1 << TOIE0;

	// start!

	kb_set_clock_0();
	kb_set_data_1();

	qinit(&event_queue_l);
	qinit(&event_queue_h);
	initialize_kb_input();
	initialize_kb_output();

	kb_state = KB_INIT;
	key_decode_state = 0;
	shiftflags = 0;
	indicator = 0;
	real_keyboard_id = 0x83ab;
	fake_keyboard_id = 0x83ab;
	break_enable = true;
	scan_enable = true;

	pspad_state = 0;
	pspad_buttons0 = 0xff;
	pspad_buttons1 = 0xff;
	pspad_reset_state();

	kb_set_clock_1();
	sei();

	put_event(EVENT_INIT);

	// event loop

	while (1) {
		main_handler();
		system_handler();
	}

	return 0;
}





